import hashlib
import json
import sys
from dataclasses import dataclass, field
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Dict, Tuple, Union, cast

import smart_open
from omegaconf import MISSING
from omegaconf.omegaconf import OmegaConf as om

from dolma.core.parallel import BaseParallelProcessor


class TrafilaturaReformatter(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore
        cls,
        queue: "Queue[Union[Tuple[int, ...], None]]",
        /,
        files: int = 0,
        documents: int = 0,
    ) -> Dict[str, int]:
        return super().increment_progressbar(queue, files=files, documents=documents)

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        documents = 0
        interval = 10_000

        destination_prefix = kwargs.get("destination_prefix", None)

        new_destination_suffix = (
            destination_path.replace(destination_prefix, "").lstrip("/").replace("/", "_") + ".gz"
        )
        destination_path = f"{destination_prefix.rstrip()}/{new_destination_suffix}"

        with smart_open.open(source_path, "rt") as source_file, smart_open.open(
            destination_path, "wt"
        ) as destination_file:
            for line in source_file:
                document = json.loads(line)
                documents += 1

                transformed = {
                    # use hash of the whole document as the id
                    "id": hashlib.md5(line.encode("utf-8")).hexdigest(),
                    "text": document["content"],
                    "source": "trafilatura",
                    "metadata": document,
                }

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

                destination_file.write(json.dumps(transformed) + "\n")

        cls.increment_progressbar(queue, files=1, documents=documents % interval)


@dataclass
class Config:
    src: str = field(default_factory=lambda: MISSING)
    dst: str = field(default_factory=lambda: MISSING)
    proc: int = field(default=1)
    debug: bool = field(default=False)


def reformat_files(config: Config):
    with TemporaryDirectory() as tempdir:
        processor = TrafilaturaReformatter(
            source_prefix=config.src,
            destination_prefix=config.dst,
            metadata_prefix=tempdir,
            num_processes=config.proc,
            debug=config.debug,
        )
        processor(destination_prefix=config.dst)


if __name__ == "__main__":
    config = om.merge(om.structured(Config), om.from_cli(sys.argv[1:]))
    reformat_files(cast(Config, config))



class DolmaError(Exception):
    """Base class for all errors"""


class DolmaFatalError(DolmaError):
    """Fatal error. Abort the entire process"""


class DolmaShardError(DolmaError):
    """Fail the shard and continue"""


class DolmaRetryableFailure(DolmaError):
    """Retry if a shard throws this error"""


class DolmaRustPipelineError(DolmaError):
    """Error raised by the rust pipeline"""


class DolmaConfigError(DolmaError):
    """Error raised while parsing config"""



from pathlib import Path
from unittest import TestCase

from dolma.tokenizer import Tokenizer

TEST_DIR = Path(__file__).parent.parent.resolve()


LLAMA_TOKENIZER = {
    "filename": f"{TEST_DIR}/data/tokenizer/llama-test-tokenizer.json",
    "bos_token_id": 1,
    "eos_token_id": 2,
    "pad_token_id": None,
}
GPT_NEO_TOKENIZER = {
    "filename": f"{TEST_DIR}/data/tokenizer/gpt-neo-test-tokenizer.json",
    "bos_token_id": None,
    "eos_token_id": 50279,
    "pad_token_id": 1,
}

TEXT_WITH_NO_NEWLINES = {
    "text": "This is a document with no newlines.",
    "llama": [1, 910, 338, 263, 1842, 411, 694, 716, 9012, 29889, 2],
    "gpt_neo": [1552, 310, 247, 3389, 342, 642, 747, 8737, 15, 50279],
}

TEXT_WITH_NEW_LINES = {
    "text": "A doc with\nnewlines.\n\nToks be the same!\n",
    "llama": [1, 319, 1574, 411, 13, 1482, 9012, 29889, 13, 13, 29911, 12117, 367, 278, 1021, 29991, 13, 2],
    "gpt_neo": [34, 5474, 342, 187, 1826, 8737, 15, 187, 187, 53, 24560, 320, 253, 1072, 2, 187, 50279],
}

TEXT_NEWLINE_START = {
    "text": "\nSimple doc with leading newline.",
    "llama": [1, 29871, 13, 15427, 1574, 411, 8236, 25899, 29889, 2],
    "gpt_neo": [187, 21595, 5474, 342, 4283, 747, 1282, 15, 50279],
}


class TestTokenizer(TestCase):
    def test_llama_process_by_paragraph(self):
        no_split_tok = Tokenizer.from_file(**LLAMA_TOKENIZER, segment_before_tokenization=False)
        split_tok = Tokenizer.from_file(**LLAMA_TOKENIZER, segment_before_tokenization=True)

        # Test that the tokenization is the when segment_before_tokenization is True or False
        # and no newlines are present; we compare with reference tokenization as well
        no_split_tokens = no_split_tok.encode(TEXT_WITH_NO_NEWLINES["text"])
        split_tokens = split_tok.encode(TEXT_WITH_NO_NEWLINES["text"])
        self.assertEqual(no_split_tokens, split_tokens)
        self.assertEqual(split_tokens, TEXT_WITH_NO_NEWLINES["llama"])

        # Test that the tokenization is the same when segment_before_tokenization is True or False
        # and the document has newlines; we compare with reference tokenization as well
        no_split_tokens = no_split_tok.encode(TEXT_WITH_NEW_LINES["text"])
        split_tokens = split_tok.encode(TEXT_WITH_NEW_LINES["text"])
        self.assertEqual(no_split_tokens, split_tokens)
        self.assertEqual(split_tokens, TEXT_WITH_NEW_LINES["llama"])

        # Test how it behaves when the document starts with a newline
        no_split_tokens = no_split_tok.encode(TEXT_NEWLINE_START["text"])
        split_tokens = split_tok.encode(TEXT_NEWLINE_START["text"])
        self.assertEqual(no_split_tokens, split_tokens)
        self.assertEqual(split_tokens, TEXT_NEWLINE_START["llama"])

    def test_gpt_neo_process_by_paragraph(self):
        no_split_tok = Tokenizer.from_file(**GPT_NEO_TOKENIZER, segment_before_tokenization=False)
        split_tok = Tokenizer.from_file(**GPT_NEO_TOKENIZER, segment_before_tokenization=True)

        # Test that the tokenization is the when segment_before_tokenization is True or False
        # and no newlines are present; we compare with reference tokenization as well
        no_split_tokens = no_split_tok.encode(TEXT_WITH_NO_NEWLINES["text"])
        split_tokens = split_tok.encode(TEXT_WITH_NO_NEWLINES["text"])
        self.assertEqual(no_split_tokens, split_tokens)
        self.assertEqual(split_tokens, TEXT_WITH_NO_NEWLINES["gpt_neo"])

        # Test that the tokenization is the same when segment_before_tokenization is True or False
        # and the document has newlines; we compare with reference tokenization as well
        no_split_tokens = no_split_tok.encode(TEXT_WITH_NEW_LINES["text"])
        split_tokens = split_tok.encode(TEXT_WITH_NEW_LINES["text"])
        self.assertEqual(no_split_tokens, split_tokens)
        self.assertEqual(split_tokens, TEXT_WITH_NEW_LINES["gpt_neo"])

        # Test how it behaves when the document starts with a newline
        no_split_tokens = no_split_tok.encode(TEXT_NEWLINE_START["text"])
        split_tokens = split_tok.encode(TEXT_NEWLINE_START["text"])
        self.assertEqual(no_split_tokens, split_tokens)
        self.assertEqual(split_tokens, TEXT_NEWLINE_START["gpt_neo"])



from dataclasses import dataclass
from pstats import SortKey
from typing import List, Optional

from rich.console import Console
from rich.table import Table

from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig, make_workdirs
from dolma.core.errors import DolmaConfigError
from dolma.core.loggers import get_logger
from dolma.core.paths import glob_path
from dolma.core.registry import TaggerRegistry
from dolma.core.runtime import create_and_run_tagger
from dolma.core.utils import import_modules


@dataclass
class ProfilerConfig:
    enable: bool = field(
        default=False,
        help="Whether to enable profiling.",
    )
    output: Optional[str] = field(
        default=None,
        help="Path to save the profiling output; if not provided, the output will be printed to stdout.",
    )
    steps: Optional[int] = field(
        default=None,
        help="List of steps to profile; if not provided, all steps will be profiled.",
    )
    sort_key: str = field(
        default="tottime",
        choices=[str(k) for k in SortKey._value2member_map_],  # pylint: disable=no-member,protected-access
        help="Sort key for the profiling output.",
    )
    lines: int = field(
        default=100,
        help="Number of rows to print in the profiling output.",
    )


@dataclass
class TaggerConfig:
    documents: List[str] = field(
        default=[],
        help="One or more document paths to process; Can be either local or S3 paths. Globs are supported.",
    )
    destination: Optional[List[str]] = field(
        default=None,
        nargs="*",
        help=(
            "Destination paths to save the outputs; should match the number of document paths. "
            "If not provided, destination will be derived from the document path."
        ),
    )
    tagger_modules: List[str] = field(
        default=[],
        help=(
            "Additional modules to import taggers from; this is useful for taggers that are not part of Dolma. "
            "Modules must be available in $PYTHONPATH or a path to module. Taggers should be registered using the "
            "@dolma.add_tagger(...) decorator."
        ),
    )
    taggers: List[str] = field(
        default=[],
        help="List of taggers to run.",
    )
    experiment: Optional[str] = field(
        default=None,
        help="Name of the experiment.",
    )
    processes: int = field(
        default=1,
        help="Number of parallel processes to use.",
    )
    ignore_existing: bool = field(
        default=False,
        help="Whether to ignore existing outputs and re-run the taggers.",
    )
    debug: bool = field(
        default=False,
        help="Whether to run in debug mode.",
    )
    profile: ProfilerConfig = field(
        default=ProfilerConfig(),
        help="Whether to run in profiling mode.",
    )
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    dryrun: bool = field(
        default=False,
        help="If true, only print the configuration and exit without running the taggers.",
    )


class TaggerCli(BaseCli):
    CONFIG = TaggerConfig
    DESCRIPTION = (
        "Tag documents or spans of documents using one or more taggers. "
        "For a list of available taggers, run `dolma list`."
    )

    @classmethod
    def run(cls, parsed_config: TaggerConfig):
        logger = get_logger("tagger")

        with make_workdirs(parsed_config.work_dir) as work_dirs:
            documents = [str(p) for p in parsed_config.documents]
            taggers = [str(p) for p in parsed_config.taggers]

            # perform some path validation to make sure we don't call the mixer with invalid config
            total_matching_documents = 0
            for document in documents:
                current_matching_documents = sum(1 for _ in glob_path(document))
                if current_matching_documents == 0:
                    # only raise a warning if no documents are found for a single path
                    logger.warning("No documents found for path %s", document)
                total_matching_documents += current_matching_documents

            if total_matching_documents == 0:
                # but raise an error if no documents are found for all paths
                raise DolmaConfigError(f"No documents found for paths {documents}.")

            print_config(parsed_config)
            if parsed_config.dryrun:
                logger.info("Exiting due to dryrun.")
                return

            create_and_run_tagger(
                documents=documents,
                destination=parsed_config.destination,
                metadata=work_dirs.output,
                taggers=taggers,
                taggers_modules=parsed_config.tagger_modules,
                ignore_existing=parsed_config.ignore_existing,
                num_processes=parsed_config.processes,
                experiment=parsed_config.experiment,
                debug=parsed_config.debug,
                profile_enable=parsed_config.profile.enable,
                profile_output=parsed_config.profile.output,
                profile_steps=parsed_config.profile.steps,
                profile_sort_key=parsed_config.profile.sort_key,
            )


@dataclass
class ListTaggerConfig:
    tagger_modules: List[str] = field(
        default=[],
        help="List of Python modules $PYTHONPATH to import custom taggers from.",
    )


class ListTaggerCli(BaseCli):
    CONFIG = ListTaggerConfig
    DESCRIPTION = "List available taggers."

    @classmethod
    def run(cls, parsed_config: ListTaggerConfig):
        # import tagger modules
        import_modules(parsed_config.tagger_modules)

        table = Table(title="dolma taggers", style="bold")
        table.add_column("name", justify="left", style="cyan")
        table.add_column("class", justify="left", style="magenta")

        for tagger_name, tagger_cls in sorted(TaggerRegistry.taggers()):
            tagger_repr = f"{tagger_cls.__module__}.{tagger_cls.__name__}"
            table.add_row(tagger_name, tagger_repr)

        console = Console()
        console.print(table)



"""

Unit tests for taggers/*.py

@kylel

"""

from unittest import TestCase

from dolma.core.data_types import Document
from dolma.taggers.c4 import C4Tagger, FasterC4Tagger


class TestC4Tagger(TestCase):
    def setUp(self):
        self.tagger = C4Tagger()

    def test_curly_braces_filter(self):
        doc = Document(source="", version="", id="", text="This is a test.")
        result = self.tagger.predict(doc=doc)
        self.assertEqual([r.to_json() for r in result.spans if r.type == "has_curly_brace"], [])

        doc = Document(source="", version="", id="", text="This is a test {")
        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "has_curly_brace"],
            [{"start": 0, "end": 16, "type": "has_curly_brace", "score": 1.0}],
        )

    def test_javascript_filter(self):
        doc = Document(source="", version="", id="", text="This is a test.")
        result = self.tagger.predict(doc=doc)
        self.assertEqual([r.to_json() for r in result.spans if r.type == "has_javascript"], [])

        doc = Document(source="", version="", id="", text="This is a test javascript")
        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "has_javascript"],
            [{"start": 0, "end": 25, "type": "has_javascript", "score": 1.0}],
        )

    def test_lorem_ipsum_filter(self):
        doc = Document(source="", version="", id="", text="This is a test.")
        result = self.tagger.predict(doc=doc)
        self.assertEqual([r.to_json() for r in result.spans if r.type == "has_lorem_ipsum"], [])

        doc = Document(source="", version="", id="", text="This is a lorem ipsum test\nmore test.")
        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "has_lorem_ipsum"],
            # the end should be 37 because lorem ipsum is detected at the document level
            [{"start": 0, "end": 37, "type": "has_lorem_ipsum", "score": 1.0}],
        )

    def test_line_ends_with_no_punctuation(self):
        doc = Document(
            source="",
            version="",
            id="",
            text="This is a test.\nIt has more;\nA trailing space! \nShould be good on this one.\nThis one is bad\n",
        )
        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "lines_with_no_ending_punctuation"],
            [
                {"start": 16, "end": 29, "type": "lines_with_no_ending_punctuation", "score": 1.0},
                {"start": 76, "end": 92, "type": "lines_with_no_ending_punctuation", "score": 1.0},
                {"start": 92, "end": 92, "type": "lines_with_no_ending_punctuation", "score": 1.0},
            ],
        )

    def test_lines_with_too_few_words(self):
        doc = Document(
            source="",
            version="",
            id="",
            text="Short!\nThis short.\nBarely above the limit!\nthis is last frontier.",
        )

        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "lines_with_too_few_words"],
            [
                {"start": 0, "end": 7, "type": "lines_with_too_few_words", "score": 1.0},
                {"start": 7, "end": 19, "type": "lines_with_too_few_words", "score": 1.0},
            ],
        )

    def test_naughty_words(self):
        doc = Document(source="", version="", id="", text="This sentence has no bad words.")
        result = self.tagger.predict(doc=doc)
        self.assertEqual([r.to_json() for r in result.spans if r.type == "has_naughty_word"], [])

        doc = Document(source="", version="", id="", text="This sentence mentions viagra as a bad word.")
        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "has_naughty_word"],
            [{"start": 0, "end": len(doc.text), "type": "has_naughty_word", "score": 1.0}],
        )

        doc = Document(
            source="", version="", id="", text="This sentence has ass, but not a bad word because of comma."
        )
        result = self.tagger.predict(doc=doc)
        self.assertEqual([r.to_json() for r in result.spans if r.type == "has_naughty_word"], [])

        doc = Document(
            source="",
            version="",
            id="",
            text="If I say strap on because we are on a plane, it's still a bad word.",
        )
        result = self.tagger.predict(doc=doc)
        self.assertEqual(
            [r.to_json() for r in result.spans if r.type == "has_naughty_word"],
            [{"start": 0, "end": len(doc.text), "type": "has_naughty_word", "score": 1.0}],
        )


class TestFasterC4Tagger(TestC4Tagger):
    def setUp(self):
        self.tagger = FasterC4Tagger()



# mypy: disable-error-code="unused-ignore"

import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from unittest import TestCase

import smart_open

from dolma.core.parallel import BaseParallelProcessor, QueueType

LOCAL_DATA = Path(__file__).parent.parent / "data"


class MockProcessor(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(cls, queue, /, cnt: int = 0):  # type: ignore[override]
        return super().increment_progressbar(queue, cnt=cnt)

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: QueueType,
        **kwargs: Any,
    ):
        with smart_open.open(source_path, "rb") as f, smart_open.open(destination_path, "wb") as g:
            g.write(f.read())
        queue.put((1,))


class TestParallel(TestCase):
    def test_base_parallel_processor(self):
        with self.assertRaises(ValueError):
            MockProcessor(source_prefix=[], destination_prefix=[], metadata_prefix=[])

        with TemporaryDirectory() as d:
            proc = MockProcessor(
                source_prefix=str(LOCAL_DATA / "expected"),
                destination_prefix=f"{d}/destination",
                metadata_prefix=f"{d}/metadata",
                ignore_existing=False,
            )
            proc()
            src = [p for p in os.listdir(LOCAL_DATA / "expected") if not p.startswith(".")]
            meta = [p.rstrip(".done.txt") for p in os.listdir(f"{d}/metadata")]
            dest = [p for p in os.listdir(f"{d}/destination") if not p.startswith(".")]
            self.assertEqual(sorted(src), sorted(meta))
            self.assertEqual(sorted(src), sorted(dest))

        with TemporaryDirectory() as d:
            proc = MockProcessor(
                source_prefix=str(LOCAL_DATA / "expected" / "*-paragraphs.*"),
                destination_prefix=f"{d}/destination",
                metadata_prefix=f"{d}/metadata",
                ignore_existing=False,
            )
            proc()
            src = [p for p in os.listdir(LOCAL_DATA / "expected") if "paragraphs" in p]
            meta = [p.rstrip(".done.txt") for p in os.listdir(f"{d}/metadata")]
            dest = [p for p in os.listdir(f"{d}/destination")]
            self.assertEqual(sorted(src), sorted(meta))
            self.assertEqual(sorted(src), sorted(dest))



"""

Filters.

@kylel, @soldni

"""
from abc import abstractmethod
from typing import List

from .data_types import (
    DocResult,
    Document,
    DocumentWithMetadata,
    InputSpec,
    InputSpecWithMetadata,
    TaggerOutputDictType,
)

# digits after the decimal point
TAGGER_SCORE_PRECISION = 5


class BaseTagger:
    FIELDS: List[str] = ["text"]

    @classmethod
    def train(cls, *args, **kwargs):
        raise RuntimeError("This tagger does not support training")

    @classmethod
    def test(cls, *args, **kwargs):
        raise RuntimeError("This tagger does not support testing")

    @property
    def defaults(self) -> List[str]:
        """Returns the default span types for this tagger.
        If not provided, no defaults are set when creating output."""
        return []

    @abstractmethod
    def predict(self, doc: Document) -> DocResult:
        raise NotImplementedError

    def group_output(self, doc_result: DocResult) -> TaggerOutputDictType:
        tagger_output: TaggerOutputDictType = {field: [] for field in self.defaults}
        for span in doc_result.spans:
            output = (span.start, span.end, round(float(span.score), TAGGER_SCORE_PRECISION))
            tagger_output.setdefault(span.type, []).append(output)
        return tagger_output

    def tag(self, row: InputSpec) -> TaggerOutputDictType:
        """Internal function that is used by the tagger to get data"""
        doc = Document.from_spec(row)
        doc_result = self.predict(doc)
        return self.group_output(doc_result)


class BaseTaggerWithMetadata(BaseTagger):
    @abstractmethod
    def predict(self, doc: DocumentWithMetadata) -> DocResult:  # type: ignore
        raise NotImplementedError

    def tag(self, row: InputSpecWithMetadata) -> TaggerOutputDictType:
        """Internal function that is used by the tagger to get data"""
        doc = DocumentWithMetadata.from_spec(row)
        doc_result = self.predict(doc)
        return self.group_output(doc_result)



"""

Filters.

@kylel, @soldni

"""
from typing import TYPE_CHECKING, Iterable, List, Tuple

import necessary
import pycld2 as cld2
import regex
from anyascii import anyascii

from ..core.data_types import DocResult, Document, Span, TextSlice
from ..core.ft_tagger import BaseFastTextTagger, Prediction
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger
from ..core.utils import split_paragraphs

with necessary.necessary("cld3", soft=True) as CLD3_AVAILABLE:
    if CLD3_AVAILABLE or TYPE_CHECKING:
        import cld3  # pyright:ignore pylint:disable=import-error


@TaggerRegistry.add("cld3_en_doc_v2")
class Cld3LanguageTagger(BaseTagger):
    def __init__(self) -> None:
        if not CLD3_AVAILABLE:
            raise ImportError(f"cld3 is not install, cannot instantiate {self.__class__.__name__}")

    def _predict_text(self, text: str) -> Tuple[str, float]:
        pred = cld3.get_language(text)  # pyright: ignore
        score = pred.probability if pred.language == "en" else 0.0
        return "en", score

    def predict(self, doc: Document) -> DocResult:
        lang, score = self._predict_text(doc.text)
        positive_span = Span(start=0, end=len(doc.text), type=lang, score=score)
        negative_span = Span(start=0, end=len(doc.text), type=f"not_{lang}", score=1.0 - score)
        return DocResult(doc=doc, spans=[positive_span, negative_span])


@TaggerRegistry.add("cld3_en_paragraph_v2")
class Cld3LanguageTaggerParagraph(Cld3LanguageTagger):
    def predict(self, doc: Document) -> DocResult:
        paragraphs = split_paragraphs(doc.text)
        spans: List[Span] = []
        for paragraph in paragraphs:
            lang, score = self._predict_text(paragraph.text)  # pyright: ignore
            positive_span = Span(start=paragraph.start, end=paragraph.end, type=lang, score=score)
            negative_span = Span(start=paragraph.start, end=paragraph.end, type=f"not_{lang}", score=1.0 - score)
            spans.extend((positive_span, negative_span))
        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("cld2_en_doc_v2")
class Cld2LanguageFilter(BaseTagger):
    RE_BAD_CHARS = regex.compile(r"[\p{Cc}\p{Cs}]+")

    def _sanitize_input(self, text: str) -> str:
        return self.RE_BAD_CHARS.sub("", text)

    def _to_ascii_input(self, text: str) -> str:
        return anyascii(text)

    def _identity_fn(self, text: str) -> str:
        return text

    def _predict_text(self, text: str) -> Tuple[str, float]:
        details = []
        is_reliable = False
        for fn in (self._identity_fn, self._to_ascii_input, self._sanitize_input):
            try:
                is_reliable, _, details = cld2.detect(fn(text))
                break
            except cld2.error:
                ...

        score = max([d[2] for d in details if d[0] == "ENGLISH" and is_reliable] or [0])
        return "en", score / 100.0

    def predict(self, doc: Document) -> DocResult:
        lang, score = self._predict_text(doc.text)
        positive_span = Span(start=0, end=len(doc.text), type=lang, score=score)
        negative_span = Span(start=0, end=len(doc.text), type=f"not_{lang}", score=1.0 - score)
        return DocResult(doc=doc, spans=[positive_span, negative_span])


@TaggerRegistry.add("cld2_en_paragraph_v2")
class Cld2LanguageFilterParagraph(Cld2LanguageFilter):
    def predict(self, doc: Document) -> DocResult:
        paragraphs = split_paragraphs(doc.text)
        spans: List[Span] = []
        for paragraph in paragraphs:
            lang, score = self._predict_text(paragraph.text)  # pyright: ignore
            positive_span = Span(start=paragraph.start, end=paragraph.end, type=lang, score=score)
            negative_span = Span(start=paragraph.start, end=paragraph.end, type=f"not_{lang}", score=1.0 - score)
            spans.extend((positive_span, negative_span))
        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("ft_lang_id_en_doc_v2")
class FastTextEnglishLanguageDocumentTagger(BaseFastTextTagger):
    MODEL_PATH = "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"

    def __init__(self):
        super().__init__(model_path=self.MODEL_PATH, model_mode=self.DOCUMENT_LEVEL_TAGGER)

    def predict_slice(self, text_slice: TextSlice) -> Iterable[Prediction]:
        pred = self.classifier.predict(text_slice.text.lower().replace("\n", " ").strip(), k=-1)
        for label, score in zip(*pred):
            if label == "__label__en":
                return Prediction(label="en", score=score), Prediction(label="not_en", score=1.0 - score)
        return Prediction(label="en", score=0.0), Prediction(label="not_en", score=1.0)


@TaggerRegistry.add("ft_lang_id_en_paragraph_v2")
class FastTextEnglishLanguageParagraphTagger(FastTextEnglishLanguageDocumentTagger):
    def __init__(self):
        BaseFastTextTagger.__init__(self, model_path=self.MODEL_PATH, model_mode=self.PARAGRAPH_LEVEL_TAGGER)


def add_global_language_score_from_slice_score(result: DocResult) -> DocResult:
    # the total document score is # of characters in each "english" span multiplied by the likelihood
    # of said span being english
    try:
        doc_en_score = sum((s.end - s.start) * s.score for s in result.spans if s.type == "en") / len(
            result.doc.text
        )
        doc_not_en_score = 1 - doc_en_score
    except ZeroDivisionError:
        doc_en_score = doc_not_en_score = 0.0

    doc_level = (
        Span(start=0, end=len(result.doc.text), type="doc_en", score=doc_en_score),
        Span(start=0, end=len(result.doc.text), type="doc_not_en", score=doc_not_en_score),
    )
    result.spans.extend(doc_level)
    return result


@TaggerRegistry.add("cld2_en_paragraph_with_doc_score_v2")
class Cld2LanguageFilterParagraphWithDocScoreTagger(Cld2LanguageFilterParagraph):
    def predict(self, doc: Document) -> DocResult:
        doc_result = super().predict(doc)
        doc_result = add_global_language_score_from_slice_score(doc_result)
        return doc_result


@TaggerRegistry.add("cld3_en_paragraph_with_doc_score_v2")
class Cld3LanguageFilterParagraphWithDocScoreTagger(Cld3LanguageTaggerParagraph):
    def predict(self, doc: Document) -> DocResult:
        doc_result = super().predict(doc)
        doc_result = add_global_language_score_from_slice_score(doc_result)
        return doc_result


@TaggerRegistry.add("ft_lang_id_en_paragraph_with_doc_score_v2")
class FastTextEnglishLanguageParagraphWithDocScoreTagger(FastTextEnglishLanguageParagraphTagger):
    def predict(self, doc: Document) -> DocResult:
        doc_result = super().predict(doc)
        doc_result = add_global_language_score_from_slice_score(doc_result)
        return doc_result


# @TaggerRegistry.add("langdetect_en_doc_v1")
# class LangdetectTagger(BaseTagger):
#     def __init__(self) -> None:
#         (factory := DetectorFactory()).load_profile(PROFILES_DIRECTORY)
#         factory.set_seed(0)
#         self.detector = factory.create()
#         super().__init__()

#     # document-level, english / not english
#     def _predict_text(self, text: str) -> Tuple[str, float]:
#         try:
#             self.detector.append(text)
#             langs = self.detector.get_probabilities()
#             score, *_ = ([lang.prob for lang in langs if lang.lang == "en"] or [0.0])
#         except LangDetectException:
#             score = 0.0
#         finally:
#             self.detector.text = ""
#             self.detector.langprob = None

#         return "en", score

#     def predict(self, doc: Document) -> DocResult:
#         lang, score = self._predict_text(doc.text)
#         positive_span = Span(start=0, end=len(doc.text), type=lang, score=score)
#         negative_span = Span(start=0, end=len(doc.text), type=f"not_{lang}", score=1.0 - score)
#         return DocResult(doc=doc, spans=[positive_span, negative_span])



import hashlib
import multiprocessing
import os
import random
import tempfile
from contextlib import ExitStack
from math import ceil, log10
from queue import Queue
from typing import Any, Dict, List, Optional

import numpy as np
from typing_extensions import TypeAlias

from ..core.loggers import get_logger
from ..core.parallel import BaseParallelProcessor, QueueType
from ..core.paths import glob_path, join_path, mkdir_p
from .data_types import TokenizerOutput
from .memmap_writer import MemmapWriter
from .tokenizer import Tokenizer, tokenize_file

TokenizedSeqsQueueType: TypeAlias = "Queue[List[TokenizerOutput]]"
PathsQueueType: TypeAlias = "Queue[str]"


class MemMapParallelWriter(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore[override]
        cls,
        queue: QueueType,
        /,
        files: int = 0,
        documents: int = 0,
        tokens: int = 0,
        memmaps: int = 0,
    ) -> Dict[str, int]:
        return super().increment_progressbar(
            queue, files=files, documents=documents, tokens=tokens, memmaps=memmaps
        )

    @classmethod
    def process_single(cls, source_path: str, destination_path: str, queue: QueueType, **kwargs: Any):
        logger = get_logger(__name__)

        max_size: int = kwargs.pop("max_size", 1024 * 1024 * 1024)
        dtype: np.dtype = np.dtype(kwargs.pop("dtype", "uint16"))
        local_shuffle: int = kwargs.pop("local_shuffle", 10_000)
        ring_size: int = kwargs.pop("ring_size", 8)

        global_source_paths = kwargs.pop("grouped_source_prefixes", None)
        if not isinstance(global_source_paths, list):
            raise RuntimeError("grouped_source_prefixes should be a list of paths")
        source_paths = global_source_paths[int(source_path)]

        tokenizer_name_or_path = kwargs.pop("tokenizer_name_or_path", None)
        if tokenizer_name_or_path is None:
            raise RuntimeError("tokenizer_name_or_path not provided")

        eos_token_id = kwargs.pop("eos_token_id", None)
        if eos_token_id is None:
            raise ValueError("eos_token_id not provided")

        pad_token_id = kwargs.pop("pad_token_id", None)
        if pad_token_id is None:
            logger.warning("pad_token_id not provided, using eos_token_id")
            pad_token_id = eos_token_id

        # this is useful for making sure the queue does not grows too much
        cpu_count = multiprocessing.cpu_count()

        documents_cnt = tokens_cnt = 0
        update_interval = 1
        mm_cnt = 0

        tokenizer = Tokenizer.from_pretrained(
            tokenizer_name_or_path,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
        )
        tokenizer_ring = []
        for _ in range(min(ring_size, len(source_paths))):
            path = source_paths.pop()
            tokenizer_ring.append(tokenize_file(tokenizer=tokenizer, path=path))

        accumulator = []

        with ExitStack() as stack:
            memwriter = stack.enter_context(
                MemmapWriter(path=destination_path + f"-{mm_cnt:05d}", dtype=dtype, max_tokens=max_size)
            )
            cls.increment_progressbar(queue, memmaps=1)

            while len(source_paths) > 0 or len(tokenizer_ring) > 0:
                for i in range(local_shuffle):
                    j = i % len(tokenizer_ring)
                    try:
                        content = next(tokenizer_ring[j])
                        accumulator.append(content)
                    except StopIteration:
                        cls.increment_progressbar(queue, files=1)
                        tokenizer_ring.pop(j)
                        if len(tokenizer_ring) == 0:
                            break
                        if len(source_paths) > 0:
                            path = source_paths.pop()
                            tokenizer_ring.append(tokenize_file(tokenizer=tokenizer, path=path))

                    # shuffle sequence order to ensure that the sequences are well mixed
                    random.shuffle(accumulator)

                    # try to write all the sequences, collect the ones that don't fit in remaining
                    remaining = memwriter.write_many(outputs=accumulator, flush=documents_cnt == 0)

                    if remaining:
                        # if we have remaining sequences, we need to close the current memwriter and open a new one
                        mm_cnt += 1
                        stack.pop_all().close()
                        memwriter = stack.enter_context(
                            MemmapWriter(
                                path=destination_path + f"-{mm_cnt:05d}",
                                dtype=np.dtype("uint16"),
                                max_tokens=max_size,
                            )
                        )
                        cls.increment_progressbar(queue, memmaps=1)

                        # finally, write the remaining sequences
                        memwriter.write_many(outputs=remaining, flush=True)

                    tokens_cnt += sum(seq.end for seq in accumulator)
                    documents_cnt += len(accumulator)

                    if documents_cnt >= update_interval:
                        cls.increment_progressbar(queue, documents=documents_cnt, tokens=tokens_cnt)
                        tokens_cnt = documents_cnt = 0

                        if queue.qsize() >= cpu_count:
                            # double the update interval if the queue is full
                            update_interval *= 2

                    accumulator = []

                memwriter.flush()

        cls.increment_progressbar(queue, documents=documents_cnt, tokens=tokens_cnt)

    def __call__(self, num_readers: Optional[int] = None, **process_single_kwargs: Any):
        """Run the processor."""

        # get all source paths; shuffle them well
        all_source_paths = [p for p in set([p for source in self.src_prefixes for p in glob_path(source)])]
        random.shuffle(all_source_paths)

        # group source paths into a number determined by num_readers or the (# sources / number of processes)
        num_readers_per_writer = float(num_readers or len(all_source_paths) / self.num_processes)
        grouped_source_prefixes = []
        i = 0.0
        while i < len(all_source_paths):
            grouped_source_prefixes.append(all_source_paths[int(i) : int(i + num_readers_per_writer)])
            i += num_readers_per_writer

        # redefine num_processes to be the number of groups
        self.num_processes = min(len(grouped_source_prefixes), self.num_processes)

        # this is a bit of a hack but: we pass indices to grouped_source_prefixes to the processors
        # so that they can load the correct source paths
        source_indices = [str(i) for i in range(len(grouped_source_prefixes))]

        # check that only one value of destination and metadata is provided
        if len(set(self.dst_prefixes)) != 1 or len(set(self.meta_prefixes)) != 1:
            raise ValueError("Only one destination and metadata should be provided.")

        # make necessary destination directories
        destination = self.dst_prefixes[0]
        mkdir_p(destination)

        # each parallel processor will write a file name like part-dddddd-dddd.npy and part-dddddd-dddd.csv.gz
        digits = int(ceil(log10(len(grouped_source_prefixes) + 1)))
        destinations = [
            join_path(None, destination, f"part-{i:0{digits}d}") for i in range(len(grouped_source_prefixes))
        ]

        # same for metadata
        metadata = self.meta_prefixes[0]
        mkdir_p(metadata)
        metadatas = [join_path(None, metadata, f"{i}.done") for i in range(len(destinations))]

        # finally run the processors
        fn = self._debug_run_all if self.debug else self._multiprocessing_run_all
        fn(
            all_source_paths=source_indices,
            all_destination_paths=destinations,
            all_metadata_paths=metadatas,
            grouped_source_prefixes=grouped_source_prefixes,
            **process_single_kwargs,
        )


def tokenize_in_parallel(
    sources: List[str],
    destination: str,
    num_writers: int = 1,
    num_readers: Optional[int] = None,
    local_shuffle: int = 10_000,
    ring_size: int = 8,
    tokenizer_name_or_path: str = "allenai/gpt-neox-olmo-dolma-v1_5",
    eos_token_id: Optional[int] = 50279,
    pad_token_id: Optional[int] = 1,
    segment_before_tokenization: bool = False,
    seed: int = 3920,
    metadata_dir: Optional[str] = None,
    max_size: int = 1024 * 1024 * 1024,
    dtype: str = "uint16",
    debug: bool = False,
):
    # variables for the nice debugging and tokenizers
    os.environ["PYTHONBREAKPOINT"] = "ipdb.set_trace"
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    # do it once so it gets cached
    Tokenizer.from_pretrained(tokenizer_name_or_path, truncate_to=None)

    # get a run hash
    run_hash = hashlib.sha256(("".join(sources) + tokenizer_name_or_path).encode("utf-8")).hexdigest()[:8]
    metadata_dir = metadata_dir or join_path(None, tempfile.gettempdir(), f"dolma-{run_hash}")

    parallel_writer = MemMapParallelWriter(
        source_prefix=sources,
        # the call action will actually get the first destination and
        # make relative paths from there. Unfortunately, BaseParallelProcessor
        # expects as many destinations as there are sources, so we employ
        # this "hack" (that is, repeating destination len(sources) times)
        # to get around that. Same thing applies to metadata_dir.
        destination_prefix=[destination for _ in sources],
        metadata_prefix=[metadata_dir for _ in sources],
        num_processes=num_writers,
        seed=seed,
        debug=debug,
    )
    parallel_writer(
        num_readers=num_readers,
        local_shuffle=local_shuffle,
        ring_size=ring_size,
        max_size=max_size,
        dtype=dtype,
        pad_token_id=pad_token_id,
        eos_token_id=eos_token_id,
        segment_docs_before_tokenization=segment_before_tokenization,
        tokenizer_name_or_path=tokenizer_name_or_path,
    )



"""

Code-related taggers.

@akshitab

"""
import logging
import re
from typing import List

import numpy as np
from necessary import necessary

from ...core.data_types import DocResult, Document, DocumentWithMetadata, Span
from ...core.registry import TaggerRegistry
from ...core.taggers import BaseTagger, BaseTaggerWithMetadata

with necessary(["detect_secrets", "bs4", "regex", "pygments"], soft=True) as CODE_DEPENDENCIES_AVAILABLE:
    if CODE_DEPENDENCIES_AVAILABLE:
        from .starcoder import get_nl_ratio
        from .utils import (
            filter_html,
            get_ext_to_lang_mapping,
            get_secrets,
            get_whitespace_regex,
        )


logger = logging.getLogger(__name__)


def check_code_dependencies() -> None:
    """Check if code dependencies are available."""
    if not CODE_DEPENDENCIES_AVAILABLE:
        raise RuntimeError("Code dependencies are not available; please run `pip install dolma[code]`")


@TaggerRegistry.add("code_secrets_v1")
class CodeSecretsTagger(BaseTagger):
    def __init__(self) -> None:
        check_code_dependencies()
        super().__init__()

    @classmethod
    def _extract_code_secrets(cls, text: str) -> List[Span]:
        secrets_spans: List[Span] = []

        text_lines = text.splitlines()
        secrets = get_secrets(text)
        for _, secret in secrets:
            line_number = secret.line_number - 1
            span = secret.secret_value
            span_line = text_lines[line_number]
            line_start = text.find(span_line)
            start = line_start + span_line.find(span or "")
            end = start + len(span or "")
            assert text[start:end] == span
            secret_type = secret.type.replace(" ", "_")
            secrets_spans.append(Span(start=start, end=end, type=f"SECRET_{secret_type}"))  # , span])

        return secrets_spans

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""
        spans = self._extract_code_secrets(doc.text)

        # document-level score
        score = self._score(text=doc.text, secrets_spans=spans)
        spans.append(Span(start=0, end=len(doc.text), type="doc", score=score))
        return DocResult(doc=doc, spans=spans)

    def _score(self, text: str, secrets_spans: List[Span]) -> float:
        try:
            score = len(secrets_spans) * 1.0 / len(text.split())
        except ZeroDivisionError:
            score = -1.0
        return score


@TaggerRegistry.add("code_copyright_comments_v1")
class CodeCopyrightTagger(BaseTagger):
    """
    Based on RedPajama code filtering.
    """

    def __init__(self):
        check_code_dependencies()
        self.cpat = re.compile("copyright", re.IGNORECASE)
        self.pat = re.compile("/\\*[^*]*\\*+(?:[^/*][^*]*\\*+)*/")

    def _extract_copyright_spans(self, text: str) -> List[Span]:
        copyright_spans: List[Span] = []

        reg = self.pat.search(text)

        if reg:
            # found one, now see if it contains "copyright", if so strip it
            span = reg.span()
            sub = text[span[0] : span[1]]
            if self.cpat.search(sub):
                copyright_spans.append(Span(start=span[0], end=span[1], type="copyright_notice", score=1.0))
            return copyright_spans

        lines = text.split("\n")
        skip = 0
        # Greedy replace any file that begins with comment block, most
        # are copyright headers
        end = 0
        for line in lines:
            if line.startswith("//") or line.startswith("#") or line.startswith("--") or not line:
                skip = skip + 1
                if not line:
                    end += 1
                else:
                    end += len(line)
            else:
                break

        if skip:
            copyright_spans.append(Span(start=0, end=end, type="comment_block", score=1.0))
        return copyright_spans

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""
        spans = self._extract_copyright_spans(doc.text)

        # document-level score
        score = self._score(text=doc.text, copyright_spans=spans)
        spans.append(Span(start=0, end=len(doc.text), type="doc", score=score))
        return DocResult(doc=doc, spans=spans)

    def _score(self, text: str, copyright_spans: List[Span]) -> float:
        try:
            if len(copyright_spans) == 0:
                score = 0.0
            else:
                span = copyright_spans[0]
                # percentage of content affected
                score = (span.end - span.start + 1) * 1.0 / len(text)
        except ZeroDivisionError:
            score = -1.0
        return score


@TaggerRegistry.add("code_redpajama_taggers_v1")
class CodeRedPajamaTaggers(BaseTagger):
    """
    Based on RedPajama code filtering.
    """

    def __init__(self):
        check_code_dependencies()
        self.whitespace_regex = get_whitespace_regex()
        super().__init__()

    def _get_num_tokens(self, text: str) -> int:
        return len(self.whitespace_regex.split(text))

    def predict(self, doc: Document) -> DocResult:
        """Main runner."""

        spans: List[Span] = []

        doc_length = len(doc.text)

        line_lengths = list(map(len, doc.text.splitlines()))

        max_line_length = max(line_lengths, default=0.0)
        avg_line_length = float(np.mean(line_lengths) if len(line_lengths) > 0 else 0.0)

        alnum_count = sum(map(lambda char: 1 if char.isalnum() else 0, doc.text))
        alnum_prop = (alnum_count / doc_length) if doc_length > 0 else 0.0

        num_tokens = self._get_num_tokens(doc.text)
        num_alpha = len([c for c in doc.text if c.isalpha()])
        alpha_token_prop = float((num_alpha / num_tokens) if num_tokens > 0 else 0.0)

        # document-level scores
        spans.append(Span(start=0, end=doc_length, type="max_line_length_doc", score=max_line_length))
        spans.append(Span(start=0, end=doc_length, type="avg_line_length_doc", score=avg_line_length))
        spans.append(Span(start=0, end=doc_length, type="alnum_prop_doc", score=alnum_prop))
        spans.append(Span(start=0, end=doc_length, type="alpha_token_prop_doc", score=alpha_token_prop))

        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("code_starcoder_taggers_v1")
class CodeStarCoderTaggers(BaseTaggerWithMetadata):
    """
    Based on StarCoder filtering.
    """

    def __init__(self) -> None:
        check_code_dependencies()
        self.ext_to_lang_mapping = get_ext_to_lang_mapping()
        super().__init__()

    def predict(self, doc: DocumentWithMetadata) -> DocResult:  # type: ignore
        spans: List[Span] = []
        doc_length = len(doc.text)

        has_xml_template = 1.0 if "<?xml version=" in doc.text[:100] else 0.0
        num_github_stars = doc.metadata.get("max_stars_count", 0) or 0

        try:
            lang = self.ext_to_lang_mapping[doc.metadata.get("ext", "-no-lang")]
            nl_ratio = get_nl_ratio(doc.text, lang)

            if lang == "html":
                code_to_text_ratio = filter_html(doc.text)
            else:
                # Not relevant for non-html code
                code_to_text_ratio = 1.0
        except:  # pylint: disable=bare-except   # noqa: E722
            nl_ratio = -1.0
            code_to_text_ratio = -1.0

        # document-level scores
        spans.append(Span(start=0, end=doc_length, type="has_xml_template_doc", score=has_xml_template))
        spans.append(Span(start=0, end=doc_length, type="num_github_stars_doc", score=float(num_github_stars)))
        spans.append(Span(start=0, end=doc_length, type="nl_ratio_doc", score=nl_ratio))
        spans.append(Span(start=0, end=doc_length, type="code_to_text_ratio_html_doc", score=code_to_text_ratio))

        return DocResult(doc=doc, spans=spans)


@TaggerRegistry.add("code_starcoder_taggers_v2")
class CodeStarCoderTaggers2(BaseTaggerWithMetadata):
    """
    Based on StarCoder filtering.
    """

    def __init__(self) -> None:
        check_code_dependencies()
        self.ext_to_lang_mapping = get_ext_to_lang_mapping()
        super().__init__()

    def predict(self, doc: DocumentWithMetadata) -> DocResult:  # type: ignore
        spans: List[Span] = []
        doc_length = len(doc.text)

        has_xml_template = 1.0 if "<?xml version=" in doc.text[:100] else 0.0
        num_github_stars = doc.metadata.get("max_stars_count", 0) or 0

        try:
            lang = self.ext_to_lang_mapping[doc.metadata.get("ext", "-no-lang")]
        except KeyError:
            lang = "-no-lang"

        if lang in ["python", "java", "javascript"]:
            code_to_comment_ratio = get_nl_ratio(doc.text, lang)
        else:
            code_to_comment_ratio = 0.5  # We use an upper and lower bound of filters; this is in the middle.

        if lang == "html":
            try:
                code_to_text_ratio = filter_html(doc.text)
            except:  # pylint: disable=bare-except   # noqa: E722
                code_to_text_ratio = -1.0
        else:
            code_to_text_ratio = 1.0

        # document-level scores
        spans.append(Span(start=0, end=doc_length, type="has_xml_template_doc", score=has_xml_template))
        spans.append(Span(start=0, end=doc_length, type="num_github_stars_doc", score=float(num_github_stars)))
        spans.append(Span(start=0, end=doc_length, type="code_to_comment_ratio_doc", score=code_to_comment_ratio))
        spans.append(Span(start=0, end=doc_length, type="code_to_text_ratio_html_doc", score=code_to_text_ratio))

        return DocResult(doc=doc, spans=spans)



"""A Dataflow script for creating DOLMA formatted pretraining data from
reddit comment and submission dumps.

For usage see README.md.

Adapted from work done at PolyAI:
https://github.com/PolyAI-LDN/conversational-datasets/tree/master/reddit
"""

import logging
from functools import partial
from collections import defaultdict
import uuid
import random
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from utils.shared_utils import (
    build_base_parser,
    positive_int,
    normalize_id,
    normalize_string,
    load_filtered_subreddit_lists,
    read_content_from_source,
    write_to_gcs,
    isodatetime_from_epoch,
    DATA_ACQUISITION_DATE
)


def parse_args(argv=None):
    parser = build_base_parser()

    parser.add_argument(
        "--parent_depth",
        type=positive_int,
        help="How many parents to consider.",
    )
    parser.add_argument(
        "--max_length",
        type=positive_int,
        default=10000,
        help="Maximum letters in content to include.",
    )
    parser.add_argument(
        "--min_length",
        type=positive_int,
        default=9,
        help="Minimum characters in content to include.",
    )
    parser.add_argument(
        "--min_conversational_length",
        type=positive_int,
        default=500,
        help="Minimum characters in content to include.",
    )
    return parser.parse_known_args(argv)


def normalize_comment(comment, max_length):

    def trim(text, max_length):
        if len(text) <= max_length:
            return text
        text = text[:max_length + 1]

        # Trim until the last two characters are the boundary between an
        # alphanumeric character, and a non-alphanumeric character.
        while len(text) > 1 and (text[-1].isalnum() == text[-2].isalnum()):
            text = text[:-1]
        return text[:-1]

    comment_score = comment.get('score', 0)
    comment = {
        'id': comment['id'],
        'thread_id': normalize_id(comment['link_id']),
        'parent_id': normalize_id(comment['parent_id']),
        'body': trim(normalize_string(comment['body']), max_length),
        'body_is_trimmed': len(comment['body']) > max_length,
        'author': normalize_string(comment['author']),
        'subreddit': normalize_string(comment['subreddit']),
        'created_utc': comment['created_utc'],
        'score': comment_score,
        'link_id': comment['link_id']
    }
    return comment

def normalize_post(post, max_length):

    body_key = "body" if "body" in post else "selftext"
    post = {
        "id": post['id'],
        "title": normalize_string(post['title']) if 'title' in post else '',
        "author": normalize_string(post['author']) if 'author' in post else '',
        "subreddit": normalize_string(post['subreddit']) if 'subreddit' in post else '',
        "subreddit_id": normalize_id(post['subreddit_id']) if 'subreddit_id' in post else '',
        "body": normalize_string(post[body_key]),
        "body_is_trimmed": len(post[body_key]) > max_length,
        "created_utc": post['created_utc'] if 'created_utc' in post else None,
        "score": post['score'] if 'score' in post else 0,
        "over_18": post['over_18'] if 'over_18' in post else True,
        "num_comments": post['num_comments'] if 'num_comments' in post else 0,
    }
    return post


def create_examples(combined_thread, parent_depth, min_length):

    def _should_skip(comment, min_length):
        if comment['body_is_trimmed']:
            return True
        if comment['body'] in {
            "[deleted]",
            "[removed]",
            "[UNICODE ENCODE ERROR]"}:
            return True
        if comment['subreddit'] in {
            "[deleted]",
            "[removed]",
            "[UNICODE ENCODE ERROR]"}:
            return True
        if len(comment['body']) < min_length:
            return True
        return False

    def build_anonymous_author_names(thread):
        author_to_anonymous_name = {}
        for comment in list(thread):
            rand_int = random.randrange(1, 10 ** 4)
            random_author_name = f'user_{str(rand_int).zfill(4)}'
            author_to_anonymous_name[comment['author']] = random_author_name
        return author_to_anonymous_name

    def assemble_convo_string(cid, thread, anonymous_author_lookup, nodes, level):
        if level == parent_depth:
            return ''
        comment = thread[cid]
        convo_string = ''
        indent = ' ' * 4
        text = comment['body']
        author = anonymous_author_lookup[comment['author']]
        convo_string += f'{indent * level }#{author}#: {text}\n\n'
        for child in sorted(nodes.get(cid, []), key=lambda x: thread[x]['score'], reverse=True):
            convo_string += assemble_convo_string(child, thread, anonymous_author_lookup, nodes, level + 1)
        return convo_string

    def order_comments(thread, anonymous_author_lookup):
        nodes, roots = defaultdict(set), set()
        for cid, comment in thread.items():
            if _should_skip(comment, 1):
                continue
            if comment['link_id'].split('_')[1] == comment['parent_id']:
                roots.add(cid)
            else:
                nodes[comment['parent_id']].add(cid)
        thread_string = ''
        for cid in sorted(roots, key=lambda x: thread[x]['score'], reverse=True):
            thread_string += assemble_convo_string(cid, thread, anonymous_author_lookup, nodes, 0)
        return thread_string

    thread = combined_thread[1][0]
    thread_post = combined_thread[1][1]

    if thread and thread_post:
        thread_post = thread_post[0]
        id_to_comment = {
            comment['id']: comment for comment in list(thread)}

        anonymous_author_lookup = build_anonymous_author_names(thread)

        if thread_post:
            preamble = '#title#: ' + thread_post['title'] + '\n\n'
            if thread_post.get('body', '') and not _should_skip(thread_post, min_length):
                preamble += '#body#: ' + thread_post['body'] + '\n\n'
        else:
            preamble = ''

        conv_str = preamble + order_comments(
            id_to_comment, anonymous_author_lookup)

        example = {
            'conversational_format': conv_str,
            'subreddit': thread_post['subreddit'],
            'thread_id': thread_post['id'],
            'created': isodatetime_from_epoch(thread_post['created_utc']),
            'added': DATA_ACQUISITION_DATE,
            'id': uuid.uuid4().hex,
        }
        if len(conv_str) < 500:
            example = None
    else:
        example = None
    yield example


def run(argv=None, comments=None, posts=None):

    args, pipeline_args = parse_args(argv)
    banned_subreddits_file = args.banned_subreddits_file
    banned_subreddits = load_filtered_subreddit_lists(banned_subreddits_file)

    pipeline_options = PipelineOptions(pipeline_args, save_main_session=True,
                                       dataflow_service_options=['enable_prime'],
                                       experiments=['enable_batch_vmr',
                                                    'enable_vertical_memory_autoscaling',
                                                    'no_use_multiple_sdk_containers'])

    p = beam.Pipeline(options=pipeline_options)

    comments = read_content_from_source(comments, p, args.input_gcs_dir_comments)

    comments |= (
        "normalize content" >> beam.Map(
            partial(normalize_comment, max_length=args.max_length)))

    posts = read_content_from_source(posts, p, args.input_gcs_dir_submissions)

    posts |= (
        "normalize posts" >> beam.Map(
            partial(normalize_post, max_length=args.max_length)))


    comment_id_to_comments = comments | (
        "Key comments by thread id" >> beam.Map(
            lambda comment: (comment['thread_id'], comment)))

    post_id_to_posts = posts | (
        "Key posts by thread id" >> beam.Map(
            lambda post: (post['id'], post)))

    examples  = (
            (comment_id_to_comments, post_id_to_posts)
            | 'Join comments and submissions' >> beam.CoGroupByKey()
            | beam.FlatMap(partial(create_examples,
                    parent_depth=args.parent_depth,
                    min_length=args.min_length,
                    min_conversation_length=args.min_conversation_length
                    ))
    )

    examples = examples | ("Filtering posts that are skipped" >>
                     beam.Filter(lambda example: example is not None))

    write_to_gcs(examples, banned_subreddits, args)

    result = p.run()
    result.wait_until_finish()


if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    run()



import multiprocessing
from typing import List, TypeVar

from cached_path import cached_path
from omegaconf.omegaconf import OmegaConf as om
from omegaconf.omegaconf import Resolver

from ..core.paths import glob_path

__all__ = ["cache", "glob", "processes"]


C = TypeVar("C", bound=Resolver)


def resolver(resolver: C) -> C:
    resolver_name = f"d.{resolver.__name__}"
    om.register_new_resolver(resolver_name, resolver, replace=True)
    return resolver


@resolver
def cache(path: str) -> str:
    return str(cached_path(path))


@resolver
def glob(path: str) -> List[str]:
    globbed = list(glob_path(path))
    assert len(globbed) > 0, f"Path {path} does not match any files"
    return globbed


@resolver
def processes(n: int = 0) -> int:
    return max(1, multiprocessing.cpu_count() - n)



import functools
import os
import re
from contextlib import ExitStack
from csv import writer
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Optional, TextIO

import numpy as np
import smart_open

from ..core.loggers import get_logger
from .data_types import MemmapMetadata, TokenizerOutput

log = get_logger(__name__)


class MemmapWriter:
    """Context manager responsible for writing, resizing, and closing / uploading a memmap file."""

    DEFAULT_MAX_TOKENS = 512 * 1024 * 1024  # 500M tokens / 1GB
    MEMMAP_EXTENSION = ".npy"
    METADATA_EXTENSION = ".csv.gz"

    def __init__(
        self,
        path: str,
        dtype: np.dtype,
        max_tokens: int = DEFAULT_MAX_TOKENS,
    ):
        """Create a new memmap file.

        Args:
            path (str): Location for the memmap file. If the path is not local, the memmap file will be
                written to a temporary file first and then uploaded to the destination.
            dtype (np.dtype): Data type for the memmap file; must be a valid numpy dtype.
            max_tokens (int, optional): Maximum number of tokens per file. Defaults to 500M tokens, which is 1GB.
        """
        base_path = re.sub(r"(\.npy?)?(\.[a-zA-Z]+)*$", "", path)
        self.memmap_path = f"{base_path}{self.MEMMAP_EXTENSION}"
        self.metadata_path = f"{base_path}{self.METADATA_EXTENSION}"
        self.dtype = dtype
        self.max_tokens = max_tokens

        self._local_memmap_path: Optional[Path] = None
        self._local_metadata_path: Optional[Path] = None
        self._written_tokens = 0
        self._memmap_file: Optional[np.memmap] = None
        self._metadata_file: Optional[TextIO] = None

    @functools.cached_property
    def metadata_writer(self):
        if self._metadata_file is None:
            raise RuntimeError("Metadata file is not open")
        return writer(self._metadata_file)

    def __len__(self) -> int:
        """Length of the memmap file in tokens that have been written."""
        return self._written_tokens

    def write(self, output: TokenizerOutput, flush: bool = False) -> bool:
        """Write a list of token IDs to the memmap file; if only a subset of the values can be written,
        return the rest.

        Args:
            values (List[int]): List of token IDs to write.
            flush (bool, optional): Whether to flush the memmap file after writing. Defaults to False.
        """

        if self._memmap_file is None:
            raise RuntimeError("MemmapFile is not open")

        if self._metadata_file is None:
            raise RuntimeError("Metadata file is not open")

        if (len(output.tokens) + self._written_tokens) >= self.max_tokens:
            # return false if the memmap file is full
            return False

        metadata = MemmapMetadata(
            id=output.id,
            src=output.src,
            loc=output.loc,
            start=self._written_tokens,
            end=self._written_tokens + output.end,
        )
        self._memmap_file[self._written_tokens : self._written_tokens + output.end] = output.tokens
        self._written_tokens += output.end

        # self._metadata_file.write(msgspec.json.encode(metadata) + b"\n")
        self.metadata_writer.writerow(metadata)

        if flush:
            self.flush()

        return True

    def write_many(self, outputs: List[TokenizerOutput], flush: bool = False) -> List[TokenizerOutput]:
        remaining: List[TokenizerOutput] = []

        for i, output in enumerate(outputs):
            if not self.write(output=output):
                remaining = outputs[i:]
                break

        if flush:
            self.flush()

        return remaining

    def flush(self):
        """Flush the memmap file."""
        if self._memmap_file is not None:
            self._memmap_file.flush()

        if self._metadata_file is not None:
            self._metadata_file.flush()

    @property
    def is_remote_path(self) -> bool:
        return re.match("[a-zA-Z0-9]+://", self.memmap_path) is not None and not self.memmap_path.startswith(
            "file://"
        )

    def __enter__(self) -> "MemmapWriter":
        """Context manager entry point. Creates the memmap file and returns self."""

        assert self._memmap_file is None and self._metadata_file is None, "MemmapFile is already open"

        if self.is_remote_path:
            with ExitStack() as stack:
                # if the destination for the memmap is not local, we need to write to a temporary file first
                _memmap_file = stack.enter_context(
                    NamedTemporaryFile(delete=False, prefix="olmo_memmap", suffix=self.MEMMAP_EXTENSION)
                )
                self._local_memmap_path = Path(_memmap_file.name)
                _metadata_file = stack.enter_context(
                    NamedTemporaryFile(delete=False, prefix="olmo_metadata", suffix=self.METADATA_EXTENSION)
                )
                self._local_metadata_path = Path(_metadata_file.name)
        else:
            self._local_memmap_path = Path(self.memmap_path)
            self._local_metadata_path = Path(self.metadata_path)
            # make sure the directory exists
            self._local_memmap_path.parent.mkdir(parents=True, exist_ok=True)
            self._local_metadata_path.parent.mkdir(parents=True, exist_ok=True)

        # these two assertions ensure type checking
        assert self._local_memmap_path is not None
        assert self._local_metadata_path is not None

        self._memmap_file = np.memmap(
            mode="w+", filename=self._local_memmap_path, dtype=self.dtype, shape=(self.max_tokens,)
        )
        self._metadata_file = smart_open.open(self._local_metadata_path, mode="wt")

        log.info(f"Created memmap file at {self._local_memmap_path} of size {self._memmap_file.nbytes:,} bytes")

        return self

    def __exit__(self, *_):
        """Context manager exit point. Closes the memmap file."""
        return self.close()

    def close(self):
        """Close the memmap file and optionally upload it to the destination (in the case of a remote path)."""
        assert self._local_memmap_path is not None, "Local Memmap path is not provided"
        assert self._local_metadata_path is not None, "Local Metadata path is not provided"
        assert self._memmap_file is not None, "Memmap file is not open"
        assert self._metadata_file is not None, "Metadata file is not open"

        try:
            # write the memmap to the destination
            self.flush()
            self._metadata_file.close()

            # we resize the memmap to the number of tokens actually written
            if self._written_tokens < self.max_tokens and self._written_tokens > 0:
                del self._memmap_file
                os.rename(self._local_memmap_path, (temp_path := self._local_memmap_path.with_suffix(".tmp")))
                new_memmap = np.memmap(
                    mode="w+", filename=self._local_memmap_path, dtype=self.dtype, shape=(self._written_tokens,)
                )
                old_memmap = np.memmap(mode="r", filename=temp_path, dtype=self.dtype, shape=(self.max_tokens,))
                new_memmap[:] = old_memmap[: self._written_tokens]
                new_memmap.flush()
                log.info(f"Resized memmap file from {old_memmap.nbytes:,} to {new_memmap.nbytes:,} bytes")
                os.remove(temp_path)

            if self.is_remote_path:
                with ExitStack() as stack:
                    f = stack.enter_context(smart_open.open(self._local_memmap_path, "rb"))
                    g = stack.enter_context(smart_open.open(self.memmap_path, mode="wb"))
                    g.write(f.read())

                    f = stack.enter_context(smart_open.open(self._local_metadata_path, "rb"))
                    g = stack.enter_context(smart_open.open(self.metadata_path, mode="wb"))
                    g.write(f.read())

                log.info(f"Written memmap file to {self.memmap_path}")
        finally:
            if self.is_remote_path:
                # delete the temporary file under any circumstances
                os.remove(self._local_memmap_path)

        # reset to none, clear cache
        self._local_memmap_path = self._memmap_file = None
        self._local_metadata_path = self._metadata_file = None

        try:
            del self.metadata_writer
        except AttributeError:
            # this is in case the metadata file was never opened
            pass



from .repetitions_taggers import (
    ParagraphRepetitionsTagger,
    RepetitionsTagger,
    TokenizerRepetitionsTagger,
)

__all__ = ["RepetitionsTagger", "ParagraphRepetitionsTagger", "TokenizerRepetitionsTagger"]



import io
import multiprocessing
import tempfile
from contextlib import ExitStack, contextmanager
from typing import (
    IO,
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Union,
)

import msgspec
import smart_open

from dolma.core.taggers import BaseTaggerWithMetadata

from .data_types import (
    InputSpec,
    InputSpecWithMetadata,
    OutputSpec,
    TaggerOutputDictType,
)
from .errors import DolmaFatalError, DolmaRetryableFailure, DolmaShardError
from .parallel import BaseParallelProcessor, QueueType
from .paths import delete_dir, join_path, make_relative, mkdir_p, split_glob, split_path
from .registry import TaggerRegistry
from .utils import import_modules, make_variable_name

# this placeholder gets used when a user has provided no experiment name, and we want to use taggers'
# names as experiment names.
EXPERIMENT_PLACEHOLDER_NAME = "_______EXPERIMENT_PLACEHOLDER_NAME_______"


def _make_paths_from_substitution(paths: List[str], find: str, replace: str) -> List[str]:
    """
    Utility function to make paths using a find/replace substitution. This is useful if you want to
    create a destination path from a source path by replacing part of the source path with something else.

    For example, if you have a source path of `current_paths = ["s3://bucket/data/documents/**.json.gz"]` and
    you want to replace `documents` with `attributes`, you can use this function to do that. by calling
    `_make_paths_from_substitution(current_paths, "documents", "attribute")`. This will return the following
    list `["s3://bucket/data/attributes"]`. Note how glob patterns are removed from the paths.
    """
    new_paths: List[str] = []
    for curr in paths:
        curr_pre_glob, _ = split_glob(curr)
        curr_prot, curr_parts = split_path(curr_pre_glob)
        find_dir_index = curr_parts.index(find)

        if not curr_pre_glob.strip():
            raise RuntimeError(f"Path '{curr}' contains a wildcard at the beginning. ")
        elif find_dir_index < 0:
            raise RuntimeError(f"Path '{curr}' does not contain a '{find}' component.")

        dst_parts = [p if i != find_dir_index else replace for i, p in enumerate(curr_parts)]
        new_paths.append(join_path(curr_prot, dst_parts))

    return new_paths


def _make_paths_from_prefix(paths: List[str], prefix: str) -> List[str]:
    """
    Utility function to make paths using a prefix. This is useful if you want to create a destination path
    from a source path by prepending a prefix to the source path.

    To create destination paths, we first find the longest common prefix among all source paths. Then, we
    we remove the prefix from each source path and prepend the new prefix to each source path. For example,
    if you have a source path of
    ```
    current_paths = [
        "s3://bucket/data/documentsA/**.json.gz",
        "s3://bucket/data/documentsB/**.json.gz",
    ]
    ```
    and you want to replace `s3://bucket/data/` with `s3://bucket/attributes/`, you can use this function
    to do that. by calling `_make_paths_from_prefix(current_paths, "s3://bucket/attributes/")`. This will
    return the following list

    ```
    [
        "s3://bucket/attributes/documentsA/",
        "s3://bucket/attributes/documentsB/",
    ]
    ```

    Note how glob patterns are removed from the paths.
    """

    new_paths: List[str] = []
    prefix_prot, prefix_path = split_path(prefix)
    _, relative_paths = make_relative(paths)

    for curr_path in relative_paths:
        base_curr_path, _ = split_glob(curr_path)
        new_paths.append(join_path(prefix_prot, prefix_path, base_curr_path))

    return new_paths


class TaggerOutputLocation(NamedTuple):
    exp: str
    name: str
    path: str


class TaggerOutputIO(NamedTuple):
    exp: str
    taggers: Set[str]
    path: str
    io: IO
    encoder: msgspec.json.Encoder

    def write(self, d: OutputSpec) -> None:
        enc = self.encoder.encode(d)
        self.io.write(enc.decode("utf-8") + "\n")


def _determine_output_paths_for_taggers(
    experiment_name: str, destination: str, taggers: Iterable[str]
) -> Dict[str, TaggerOutputLocation]:
    """Utility function to derive the paths to which taggers output should be written.

    If experiment_name is the placeholder name, then the name of each tagger will be used as part of the
    destination path. Otherwise, the destination path will be used for all taggers."""

    if experiment_name == EXPERIMENT_PLACEHOLDER_NAME:
        return {
            tagger_name: TaggerOutputLocation(
                exp=make_variable_name(tagger_name),
                name=make_variable_name(tagger_name),
                path=destination.replace(EXPERIMENT_PLACEHOLDER_NAME, tagger_name),
            )
            for tagger_name in taggers
        }
    else:
        return {
            tagger_name: TaggerOutputLocation(
                exp=make_variable_name(experiment_name), name=make_variable_name(tagger_name), path=destination
            )
            for tagger_name in taggers
        }


@contextmanager
def _make_output_streams(
    taggers_paths: Dict[str, TaggerOutputLocation], **open_kwargs: Any
) -> Generator[Dict[str, TaggerOutputIO], None, None]:
    """Utility function to open paths for taggers.

    It is designed NOT to open duplicate paths if multiple taggers are writing to the same file.
    """
    # keep track of the paths that have been opened
    opened: Dict[str, TaggerOutputIO] = {}

    with ExitStack() as stack:
        for key, loc in taggers_paths.items():
            if loc.path not in opened:
                # make sure the parent directory exists
                prot, path = split_path(loc.path)
                parent = join_path(prot, path[:-1])
                mkdir_p(parent)

                # open a new file and create a new encoder
                io = stack.enter_context(smart_open.open(loc.path, **open_kwargs))
                encoder = msgspec.json.Encoder()
                opened[loc.path] = TaggerOutputIO(
                    exp=loc.exp, taggers=set(), path=loc.path, io=io, encoder=encoder
                )

            # keep track of which taggers are writing to this paths
            opened[loc.path].taggers.add(key)

        yield opened


@contextmanager
def _write_sample_to_streams(
    taggers_paths: Dict[str, TaggerOutputLocation],
    output_streams: Dict[str, TaggerOutputIO],
    row: InputSpec,
) -> Generator[Dict[str, TaggerOutputDictType], None, None]:
    """Utility function to write a sample to the output streams; yields a dictionary that should be used
    to collect the output of each tagger."""

    samples_collectors: Dict[str, TaggerOutputDictType] = {}
    yield samples_collectors

    attributes_by_stream: Dict[str, TaggerOutputDictType] = {}
    for tagger_name, tagger_data in samples_collectors.items():
        tagger_output = taggers_paths[tagger_name]
        for tagger_key, tagger_value in tagger_data.items():
            tagger_key = f"{tagger_output.exp}__{tagger_output.name}__{make_variable_name(tagger_key)}"
            attributes_by_stream.setdefault(tagger_output.path, {})[tagger_key] = tagger_value

    for stream_path, attributes in attributes_by_stream.items():
        output = OutputSpec(source=row.source, id=row.id, attributes=attributes)
        output_streams[stream_path].write(output)


class TaggerProcessor(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore
        cls,
        queue: QueueType,  # queue must be the first argument, and it should be a positional-only argument
        /,
        files: int = 0,
        documents: int = 0,
    ) -> Dict[str, int]:
        """We override this method to specify which units we want to keep track of in a progress bar.
        Specifically, we keep track of files and documents in this example. Their default value must be zero."""

        # we call the super method to increment the progress bar
        return super().increment_progressbar(queue, files=files, documents=documents)

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: QueueType,
        **kwargs,
    ):
        """Lets count run the taggers! We will use the destination path to save each tagger output."""
        # import tagger modules
        taggers_modules = kwargs.get("taggers_modules", None)
        if taggers_modules is not None:
            import_modules(taggers_modules)

        # get names of taggers
        taggers_names = kwargs.get("taggers_names", None)
        if taggers_names is None:
            raise RuntimeError("Taggers not in kwargs, this is a bug! Please report it.")
        elif not isinstance(taggers_names, list) or not all(isinstance(t, str) for t in taggers_names):
            raise RuntimeError("Taggers are in the wrong format, this is a bug! Please report it.")
        taggers = {make_variable_name(t): TaggerRegistry.get(t)() for t in taggers_names}

        # get name of experiment
        if (experiment_name := kwargs.get("experiment_name", None)) is None:
            raise RuntimeError("Experiment name not in kwargs, this is a bug! Please report it.")

        # this is the dictionary that will hold the output of each tagger
        taggers_paths = _determine_output_paths_for_taggers(
            experiment_name=experiment_name, destination=destination_path, taggers=taggers
        )

        # skip on failure
        skip_on_failure = kwargs.get("skip_on_failure", False)

        # maximum numbers of lines to process
        steps: Union[int, None] = kwargs.get("steps", None)

        # interval at which to update the progress bar; will double if it gets
        # too full
        update_interval = 1

        # running document count; gets reset every time we update the progress bar
        docs_cnt = 0

        # total number of documents processed
        total_docs_cnt = 0

        # creating dedicated decoder speeds up the process
        # if any of the taggers require metadata, we use a decoder that can handle it
        # otherwise, we use a decoder that does not parse metadata, which is faster
        if any(isinstance(tagger, BaseTaggerWithMetadata) for tagger in taggers.values()):
            decoder = msgspec.json.Decoder(InputSpecWithMetadata)
        else:
            decoder = msgspec.json.Decoder(InputSpec)

        with ExitStack() as stack:
            in_stream = stack.enter_context(smart_open.open(source_path, "rt", encoding="utf-8"))
            output_streams = stack.enter_context(
                _make_output_streams(taggers_paths=taggers_paths, mode="wt", encoding="utf-8")
            )
            try:
                for raw in in_stream:
                    row = decoder.decode(raw)

                    with _write_sample_to_streams(
                        taggers_paths=taggers_paths,
                        output_streams=output_streams,
                        row=row,
                    ) as samples_collectors:
                        # we run the taggers; the context manager will write the output to the output streams
                        for tagger_name, tagger in taggers.items():
                            samples_collectors[tagger_name] = tagger.tag(row)

                    # increment the number of documents processed so far
                    docs_cnt += 1
                    total_docs_cnt += 1

                    if steps is not None and total_docs_cnt >= steps:
                        # if we have reached the maximum number of steps, we break
                        break

                    if docs_cnt % update_interval == 0:
                        # update the progress bar every 1000 documents to prevent
                        # buffering
                        cls.increment_progressbar(queue, documents=docs_cnt)
                        docs_cnt = 0

                        if queue.qsize() >= multiprocessing.cpu_count():
                            # double the update interval if the queue is full
                            update_interval *= 2

            except Exception as exp:
                # handle any exception that might have occurred
                msg = f"Failed to process {source_path} due to {exp.__class__.__name__}: {' '.join(exp.args)}"
                if exp.__class__.__name__ == "IncompleteReadError":
                    # Intermittent error that occurs when reading from S3
                    raise DolmaRetryableFailure(msg) from exp
                else:
                    if skip_on_failure:
                        raise DolmaShardError(msg) from exp
                    else:
                        raise DolmaFatalError(msg) from exp

        # increment the files progress bar
        cls.increment_progressbar(queue, files=1, documents=docs_cnt)


@contextmanager
def profiler(
    output: Optional[str] = None,
    sort_key: str = "tottime",
    lines: int = 100,
) -> Generator[None, None, None]:
    import cProfile
    import pstats

    profile = cProfile.Profile()
    profile.enable()
    yield
    profile.disable()

    with ExitStack() as stack:
        output_stream = io.StringIO() if output is None else stack.enter_context(smart_open.open(output, "w"))
        ps = pstats.Stats(profile, stream=output_stream).sort_stats(sort_key)
        ps.print_stats(lines)


def create_and_run_tagger(
    documents: List[str],
    taggers: List[str],
    taggers_modules: Optional[List[str]] = None,
    experiment: Optional[str] = None,
    destination: Union[None, str, List[str]] = None,
    metadata: Union[None, str, List[str]] = None,
    debug: bool = False,
    seed: int = 0,
    ignore_existing: bool = False,
    skip_on_failure: bool = False,
    retries_on_error: int = 0,
    num_processes: int = 1,
    profile_enable: bool = False,
    profile_output: Optional[str] = None,
    profile_steps: Optional[int] = None,
    profile_sort_key: str = "tottime",
    profile_lines: int = 100,
):
    """This function creates a tagger and runs it on a list of documents.

    Args:
        documents (List[str]): List of documents to run the taggers on. Each element of the list is a path to
            a file containing documents in json lines format, or a glob pattern that matches such files.
        taggers (List[str]): List of taggers to run. Each element of the list is the name of a tagger.
        experiment (str, optional): The name of the experiment. This will be used to prefix the names of the
            attributes, as well as the name of the directory where the outputs will be saved in `destination`.
            If not provided, the name of each tagger will be used as the experiment name.
        destination (Union[None, str, List[str]], optional): The path where the outputs will be saved. If
            `None`, the outputs will be saved in a directory parallel to the directory containing the
            documents, with the same name as `experiment`. If a string, paths corresponding to each element
            of `documents` will be created by determining a relative path from the directory containing the
            documents.
        metadata (Union[None, str, List[str]], optional): Location where to save metadata that keeps track of
            which documents have been processed. If `None`, the metadata will be saved in a temporary directory.
        debug (bool, optional): Whether to run in debug mode. Defaults to False.
        seed (int, optional): The seed to use for the random number generator. Defaults to 0.
        ignore_existing (bool, optional): Whether to ignore existing outputs and re-run the taggers.
            Defaults to False.
        skip_on_failure (bool, optional): Whether to skip a document if it fails to process. Defaults to False.
        retries_on_error (int, optional): Number of times to retry processing a document if it fails.
            Defaults to 0 (fail immediately)
        num_processes (int, optional): Number of processes to use. Defaults to 1.
        profile_enable (bool, optional): Whether to enable profiling. Defaults to False.
        profile_output (Optional[str], optional): Path to save the profiling output; if not provided, the
            output will be printed to stdout. Defaults to None.
        profile_steps (Optional[int], optional): Number of steps to profile; if not provided, all steps will
            be profiled. Defaults to None.
        profile_sort_key (str, optional): Sort key for the profiling output. Defaults to 'tottime'.
    """

    # use placeholder experiment name if none is provided; raise an error if the placeholder name is used
    if experiment == EXPERIMENT_PLACEHOLDER_NAME:
        raise RuntimeError(f"Experiment name cannot be {EXPERIMENT_PLACEHOLDER_NAME}; reserved for internal use.")
    elif experiment is None:
        experiment = EXPERIMENT_PLACEHOLDER_NAME

    if destination is None:
        try:
            destination = _make_paths_from_substitution(documents, "documents", f"attributes/{experiment}")
        except Exception as exp:
            raise RuntimeError("Could not make destination paths from documents paths") from exp
    elif isinstance(destination, str):
        try:
            destination = _make_paths_from_prefix(documents, join_path(None, destination, experiment))
        except Exception as exp:
            raise RuntimeError(f"Could not make destination paths from prefix {destination}") from exp

    metadata = metadata or tempfile.mkdtemp()
    if isinstance(metadata, str):
        try:
            metadata = _make_paths_from_prefix(documents, metadata)
        except Exception as exp:
            raise RuntimeError(f"Could not make metadata paths from prefix {metadata}") from exp

        tagger = TaggerProcessor(
            source_prefix=documents,
            destination_prefix=destination,
            metadata_prefix=metadata,
            debug=debug or profile_enable,  # if profile is true, debug must be true
            seed=seed,
            ignore_existing=ignore_existing,
            retries_on_error=retries_on_error,
            num_processes=num_processes,
        )

        with ExitStack() as stack:
            if profile_enable:
                # start profiling
                stack.enter_context(
                    profiler(output=profile_output, sort_key=profile_sort_key, lines=profile_lines)
                )

            tagger(
                experiment_name=experiment,
                taggers_names=taggers,
                taggers_modules=taggers_modules,
                skip_on_failure=skip_on_failure,
                steps=profile_steps,
            )

        for path in destination:
            # remove any placeholder directories after computation is done
            if path.endswith(EXPERIMENT_PLACEHOLDER_NAME):
                delete_dir(path, ignore_missing=True)



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from dolma import mixer
from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig, make_workdirs
from dolma.core.errors import DolmaConfigError
from dolma.core.loggers import get_logger
from dolma.core.paths import glob_path


@dataclass
class StreamOutputConfig:
    path: str = field(help="Path where to write the mixed documents to. Required.")
    max_size_in_bytes: int = field(
        default=2 * 2**30, help="Maximum size of the output file in bytes. Defaults to 2GB."
    )
    discard_fields: List[str] = field(default=[], help="List of fields to discard from the output documents.")


@dataclass
class FilterConfig:
    include: List[str] = field(default=[], help="JSONPath expressions to include documents")
    exclude: List[str] = field(default=[], help="JSONPath expressions to exclude documents")


@dataclass
class SpanReplacementConfig:
    span: str = field(help="JSONPath expression for the span to replace")
    min_score: float = field(default=0.5, help="Minimum score for the span to be replaced")
    replacement: str = field(default="", help="Replacement for the span")


@dataclass
class StreamConfig:
    name: str = field(help="Name of the stream. Required.")
    documents: List[str] = field(default=[], help="Paths to the documents to be mixed. Required.")
    output: StreamOutputConfig = field(
        default=StreamOutputConfig(), help="Configuration for the output of the stream."
    )
    attributes: List[str] = field(default=[], help="List of attributes files to used for mixing.")
    filter: Optional[FilterConfig] = field(  # pyright: ignore
        default=None, help="Configuration for filtering documents."
    )
    span_replacement: List[SpanReplacementConfig] = field(default=[], help="Configuration for replacing spans.")


@dataclass
class MixerConfig:
    streams: List[StreamConfig] = field(default=[], help="List configurations of streams to be mixed")
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    processes: int = field(default=1, help="Number of processes to use for mixing. By default 1 process is used.")
    dryrun: bool = field(
        default=False,
        help="If true, only print the configuration and exit without running the mixer.",
    )


class MixerCli(BaseCli):
    CONFIG = MixerConfig
    DESCRIPTION = "Mix documents from multiple streams."

    @classmethod
    def run(cls, parsed_config: MixerConfig):
        logger = get_logger("mixer")

        with make_workdirs(parsed_config.work_dir) as work_dirs:
            dict_config: Dict[str, Any] = {
                "work_dir": {"input": str(work_dirs.input), "output": str(work_dirs.output)},
                "processes": int(parsed_config.processes),
                "streams": [],
            }

            for stream_config in parsed_config.streams:
                stream_config_dict: Dict[str, Any] = {}

                if stream_config.filter is not None:
                    if not stream_config.filter.include and not stream_config.filter.exclude:
                        raise DolmaConfigError("Either `include` or `exclude` must be specified for filter")

                    stream_config_dict["filter"] = {
                        "include": [str(i) for i in stream_config.filter.include],
                        "exclude": [str(i) for i in stream_config.filter.exclude],
                    }

                for span_replacement in stream_config.span_replacement:
                    stream_config_dict.setdefault("span_replacement", []).append(
                        {
                            "span": str(span_replacement.span),
                            "min_score": float(span_replacement.min_score),
                            "replacement": str(span_replacement.replacement),
                        }
                    )

                if "span_replacement" not in stream_config_dict and "filter" not in stream_config_dict:
                    raise DolmaConfigError("Either `filter` or `span_replacement` must be specified")

                # perform some path validation to make sure we don't call the mixer with invalid config
                total_matching_documents = 0
                for document in stream_config.documents:
                    if document.count("*") > 1:
                        raise DolmaConfigError("Only one wildcard is allowed in the document path")

                    current_matching_documents = sum(1 for _ in glob_path(document))
                    if current_matching_documents == 0:
                        # only raise a warning if no documents are found for a single path
                        logger.warning("No documents found for path %s", document)
                    total_matching_documents += current_matching_documents

                if total_matching_documents == 0:
                    # but raise an error if no documents are found for all paths
                    raise DolmaConfigError(f"No documents found for the paths for {stream_config.name} config.")

                # populate the stream config dict
                stream_config_dict["name"] = stream_config.name
                stream_config_dict["documents"] = [str(d) for d in stream_config.documents]
                stream_config_dict["attributes"] = [str(a) for a in list(stream_config.attributes)]
                stream_config_dict["output"] = {
                    "path": str(stream_config.output.path),
                    "max_size_in_bytes": int(stream_config.output.max_size_in_bytes),
                }

                if stream_config.output.discard_fields:
                    stream_config_dict["output"]["discard_fields"] = [
                        str(f) for f in stream_config.output.discard_fields
                    ]

                if len(stream_config_dict["documents"]) == 0:
                    raise ValueError("No documents to mix")

                dict_config["streams"].append(stream_config_dict)

            if len(dict_config["streams"]) == 0:
                raise DolmaConfigError("No streams to mix")

            print_config(dict_config)
            if parsed_config.dryrun:
                logger.info("Exiting due to dryrun.")
                return

            mixer(dict_config)



import json
import multiprocessing
from argparse import ArgumentParser
from contextlib import ExitStack
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Tuple, Union

import smart_open

from dolma.core.parallel import BaseParallelProcessor


class RemoveEmptyDocumentsProcessor(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(
        cls,
        queue: "Queue[Union[Tuple[int, ...], None]]",
        /,
        files: int = 0,
        read_docs: int = 0,
        written_docs: int = 0,
    ):
        return super().increment_progressbar(queue, files=files, read_docs=read_docs, written_docs=written_docs)

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: Queue,
        **kwargs: Any,
    ):
        """
        This method is called for each file. It reads the file
        line by line, and writes to the destination file only
        if the document is not empty.
        """

        update_every_n_lines = 1
        read_docs = written_docs = 0

        with ExitStack() as stack:
            # open source and destination files
            source_file = stack.enter_context(smart_open.open(source_path, "rt"))
            destination_file = stack.enter_context(smart_open.open(destination_path, "wt"))
            for ln in source_file:
                # we first load the json document
                document = json.loads(ln)
                read_docs += 1

                # we check if the document is
                # empty, and if it is, we skip it
                if document["text"].strip() == "":
                    continue

                # if the document is not empty,
                # we write it to output
                destination_file.write(ln)
                written_docs += 1

                # we update the progress bar every
                # update_every_n_lines
                if read_docs >= update_every_n_lines:
                    cls.increment_progressbar(
                        queue,
                        read_docs=read_docs,
                        written_docs=written_docs,
                    )
                    read_docs = written_docs = 0

                    if queue.qsize() >= multiprocessing.cpu_count():
                        # double the update interval if the queue is full
                        update_every_n_lines *= 2

            # we update the progress bar one last time
            cls.increment_progressbar(
                queue,
                files=1,
                read_docs=read_docs,
                written_docs=written_docs,
            )


def parse_args():
    ag = ArgumentParser()
    ag.add_argument("-s", "--source-prefix", type=str, required=True)
    ag.add_argument("-d", "--destination-prefix", type=str, required=True)
    ag.add_argument("-n", "--num-processes", type=int, default=1)
    ag.add_argument("-u", "--debug", action="store_true")
    ag.add_argument("-t", "--temp-dir", type=str, default=None)
    return ag.parse_args()


def main():
    args = parse_args()

    with TemporaryDirectory(dir=args.temp_dir) as tmpdir:
        # create the processor
        processor = RemoveEmptyDocumentsProcessor(
            source_prefix=args.source_prefix,
            destination_prefix=args.destination_prefix,
            metadata_prefix=tmpdir,
            num_processes=args.num_processes,
            debug=args.debug,
        )

        # run the processor
        processor()


if __name__ == "__main__":
    main()
